/*
 * Created on 24.09.2004
 *
 * To change the template for this generated file go to
 * Window&gt;Preferences&gt;Java&gt;Code Generation&gt;Code and Comments
 */


package gnu.java.beans.decoder;

import java.beans.ExceptionListener;

import org.xml.sax.Attributes;

/** An ObjectHandler parses the <object> tag and thereby create various
 * Context implementations.
 *
 * @author Robert Schuster
 *
 */
public class ObjectHandler extends AbstractElementHandler
{
  /**
   * XXX: Can all results be stored with an object id?
   *
   *
   * @param PersistenceParser
   */
  ObjectHandler(ElementHandler parent)
  {
    super(parent, true);
  }

  protected Context startElement(Attributes attributes, ExceptionListener exceptionListener)
    throws AssemblyException
  {
    String className = attributes.getValue("class");
    String methodName = attributes.getValue("method");
    String fieldName = attributes.getValue("field");
    String index = attributes.getValue("index");
    String propertyName = attributes.getValue("property");
    String id = attributes.getValue("id");
    String idRef = attributes.getValue("idref");

    /* first check if we just want to access an existing object (idref present)
     *
     * note: <object idref="foo" method="bar"/> is not valid to call method "bar"
     * on the object with id "foo". Instead this should return the object "foo"
     * itself. The right way to this is:
     * <object idref="foo">
     *         <object method="bar"/>
     * </object>
     *
     * This means that if idref is present class, method, field, index and
     * property are obsolete.
     */
    if (idRef != null)
      // reactivates an existing object and giving it another name if id exists
      return new ObjectContext(id, getObject(idRef));

    // decides whether we are in a static (className present) or dynamic context
    if (className != null)
      {
	try
	  {
	    Class klass = instantiateClass(className);

	    // class name exists which means that we are in a static context.
	    // so we may want to ...
	    // access a static field if the fieldName exists
	    if (fieldName != null)
	      {
		try
		  {
		    return new ObjectContext(id,
		                             klass.getField(fieldName).get(null));
		  }
		catch (NoSuchFieldException nsfe)
		  {
		    throw new AssemblyException(nsfe);
		  }
		catch (IllegalAccessException iae)
		  {
		    throw new AssemblyException(iae);
		  }
	      }

	    // (falling through is important!)
	    // run a constructor if methodName is "new" or null
	    if (methodName == null || methodName.equals("new"))
	      return new ConstructorContext(id, klass);

	    // (falling through is important!)
	    // run a static method on the given class (if methodName exists, which is implied already) 
	    return new StaticMethodContext(id, klass, methodName);
	    // XXX: should fail if unexpected attributes are present?
	  }
	catch (ClassNotFoundException cnfe)
	  {
	    throw new AssemblyException(cnfe);
	  }
      }
    else
      {
	// className does not exist which means we are in the context of
	// some object and want to ...
	// access the get(int index) method if index != null
	if (index != null)
	  {
	    try
	      {
		// Note: http://java.sun.com/products/jfc/tsc/articles/persistence3/ says
		// that <void index="4"/> will make up a get()-call. But this is wrong because
		// <void/> tags never return values (to the surrounding context)
		return new IndexContext(id, Integer.parseInt(index));
	      }
	    catch (NumberFormatException nfe)
	      {
		throw new AssemblyException(nfe);
	      }
	  }

	// access a method if methodName exists
	if (methodName != null)
	  return new MethodContext(id, methodName);

	// (falling through is important!)
	// access a property if a propertyName exists
	if (propertyName != null && propertyName.length() > 0)
	  // this is reported as an ordinary method access where the propertyName is
	  // converted into a 'getter'-method name: convert first character of property name
	  // to upper case and prepend 'get'
	  // Note: This will be a getter-method because the <object> tag implies that a return
	  // value is expected.
	  return new PropertyContext(id, propertyName);
      }

    throw new AssemblyException(new IllegalArgumentException("Wrong or missing attributes for <object> tag."));
  }
}
